/* ARP handling for ATM networks */
#include "otto_platform.h"

/* a type 19 ATM ATP packet */
struct arp19 {
	struct arphdr hdr;
#define oar_hrd hdr.ar_hrd	/* format of hardware address */
#define ARP19	19

#define oar_pro hdr.ar_pro	/* format of protocol address */
#define oar_shtl hdr.ar_hln	/* length/type of source ATM number */
#define oar_sstl hdr.ar_pln	/* length/type of source ATM subaddress */
#define oar_op hdr.ar_op	/* operation code */
	u_char oar_spln;	/* length of source protocol address */
	u_char oar_thtl;	/* length/type of target ATM number */
	u_char oar_tstl;	/* length/type of target ATM subaddress */
	u_char oar_tpln;	/* length of target protocol address */
/* The remaining fields are variable in size:
	u_char oar_sha[];  (oar_shtl bytes)	source ATM number
	u_char oar_ssa[];  (oar_sstl bytes)	source ATM subaddress
	u_char oar_spa[];  (oar_spln bytes)	source protocol address
	u_char oar_tha[];  (oar_thtl bytes)	target ATM number
	u_char oar_tsa[];  (oar_tstl bytes)	target ATM subaddress
	u_char oar_tpa[];  (oar_tpln bytes)	target protocol address
 */
};

#define ARP19_E164 0x40
#define ARP19_IS_E164(x) ((x) & ARP19_E164)
#define ARP19_LEN(x)	((x) & 0x3f)
#define INARP_REQUEST 8
#define INARP_REPLY   9
#define ARP_NAK       10

/* a SARP packet */
struct sarp {
	struct arphdr hdr;
#define sar_hrd hdr.ar_hrd	/* format of hardware address */
#define SARP 21
#define ANETTYPE_ETHER 1

#define sar_pro hdr.ar_pro	/* format of protocol address */
#define sar_hln hdr.ar_hln	/* length of hardware address */
#define sar_pln hdr.ar_pln	/* length of protocol address */
#define sar_op hdr.ar_op	/* operation code */
/* The remaining fields are variable in size:
        u_char sar_sha[];  (sar_hln bytes)     source hardware address
        u_char sar_spa[];  (sar_pln bytes)     source protocol address
        u_char sar_tha[];  (sar_hln bytes)     target hardware address
        u_char sar_tpa[];  (sar_pln bytes)     target protocol address
 */
};


static int myipaddr (struct otto_softc *sc, caddr_t p)
{
        struct ifnet *ifp = &sc->is_if;
        struct ifaddr *ifa;
	struct sockaddr_in *sin;

	for (ifa = ifp->if_addrlist; ifa; ifa = ifa->ifa_next) {
#ifdef __osf__
		if (ifa->ifa_addr->sa_family == AF_INET) {
			sin = (struct sockaddr_in *) ifa->ifa_addr;
#else
		if (ifa->ifa_addr.sa_family == AF_INET) {
			sin = (struct sockaddr_in *) &ifa->ifa_addr;
#endif
			bcopy ((caddr_t) &sin->sin_addr, p, 4);
			return (1);
		}
        }
        return (0);
}

/* Send a SARP request (reply=0) or reply (reply=1) on VC vc to reach
   uid "uid".  The target 
   The targer vc and target uid are tvc and tuid.
   Called with softc lock held. */
void otto_arp_sendsarp (struct otto_softc *sc, unsigned vc, unsigned tvc,
	unsigned char *uid, unsigned char *tuid, unsigned reply)
{
	unsigned msgbuf[32];
	struct sarp *nsarp = (struct sarp *) msgbuf;
	unsigned len = sizeof (struct sarp) + 20;
	unsigned homevc;
	caddr_t np;
	struct sockaddr dst;
	struct ether_header *eptr = (struct ether_header *) dst.sa_data;

	if (sc->data.snmp.pkt.homeVCI == 0) {
		return;
	}

	OTTO_STATS_INC (&sc->data, sendsarp);

	nsarp->sar_hrd = htons ((u_short) SARP);
	nsarp->sar_pro = htons ((u_short) ANETTYPE_ETHER);
	nsarp->sar_hln = 4;
	nsarp->sar_pln = 6;
	nsarp->sar_op = htons ((u_short) (reply?ARPOP_REPLY:ARPOP_REQUEST));

	np = (caddr_t) &nsarp[1];
	homevc = htonl (sc->data.snmp.pkt.homeVCI);
	bcopy ((caddr_t) &homevc, &np[0], nsarp->sar_hln);
	bcopy ((caddr_t) sc->is_addr, &np[nsarp->sar_hln],
		nsarp->sar_pln);
	tvc = htonl (tvc);
	bcopy ((caddr_t) &tvc, &np[nsarp->sar_hln+nsarp->sar_pln],
		nsarp->sar_hln);
	bcopy ((caddr_t) tuid, &np[nsarp->sar_hln*2+nsarp->sar_pln],
		nsarp->sar_pln);

	dst.sa_family = AF_UNSPEC;
	bcopy ((caddr_t)uid, (caddr_t)eptr->ether_dhost, 6);
	eptr->ether_type = ETHERTYPE_ARP;

	otto_simplesendpkt (sc, vc, msgbuf, len, &dst);
}

/* process an ARP packet in bytes p[0,...,len-1] that was received on vc "vc"
   on controller "sc".   It arrived with Ethernet header *eptr
   (the type field has been byte-swapped to host order).
   Return 1 if the packet was consumed, and 0 otherwise.
   Called with softc_lock not held. */
int otto_arp_rcv (struct otto_softc *sc, caddr_t p, int len, unsigned vc,
	struct ether_header *eptr)
{
	struct arp19 *arp19 = (struct arp19 *) p;
	struct arp19 *newarp19;
	struct sarp *sarp = (struct sarp *) p;
	struct sarp *nsarp;
	unsigned htype;
	unsigned ptype;
	unsigned op;
	unsigned lenoffs[7];	/* lengths/offsets of variable length fields */
	unsigned i;
	unsigned sum;
	unsigned addr[OTTO_MAP_NADDR];
	unsigned svc;
	unsigned msgbuf[64];
	unsigned nlen;
	caddr_t np;
	struct sockaddr dst;

	if (vc == 0 || vc >= OTTO_MAXVC) {
		printf ("otto_arp_rcv: bad vc %u\n", vc);
		panic ("otto_arp_rcv: badvc");
	}
	if (len < sizeof (struct arphdr)) { /* ARP pkt is too short */
		return (1);
	}
	htype = ntohs ((u_short) arp19->oar_hrd);

	if (htype == ARP19) {
		if (len < sizeof (struct arp19)) {
			return (1);
		}
		lenoffs[0] = sizeof (struct arp19);
		lenoffs[1] = ARP19_LEN (arp19->oar_shtl);
		lenoffs[2] = ARP19_LEN (arp19->oar_sstl);
		lenoffs[3] = arp19->oar_spln;
		lenoffs[4] = ARP19_LEN (arp19->oar_thtl);
		lenoffs[5] = ARP19_LEN (arp19->oar_tstl);
		lenoffs[6] = arp19->oar_tpln;
		sum = 0;
		for (i = 0; i != 7; i++) {
			sum += lenoffs[i];
			lenoffs[i] = sum;
		}
		if (lenoffs[6] > len) {
			return (1);
		}
		ptype = ntohs ((u_short) arp19->oar_pro);
		if (ptype != ETHERTYPE_IP) {
			return (1);
		}

		op = ntohs ((u_short) arp19->oar_op);
		switch (op) {
		case ARPOP_REQUEST:
			break;	

		case ARPOP_REPLY:
			break;	

		case INARP_REQUEST:
			newarp19 = (struct arp19 *) msgbuf;
			nlen = sizeof (struct arp19);
			newarp19->oar_hrd = htons ((u_short) ARP19);
			newarp19->oar_pro = htons ((u_short) ETHERTYPE_IP);
			newarp19->oar_shtl = 0;
			newarp19->oar_sstl = ARP19_E164;
			newarp19->oar_op = htons ((u_short) INARP_REPLY);
			newarp19->oar_spln = 4;
			newarp19->oar_thtl = arp19->oar_shtl;
			newarp19->oar_tstl = arp19->oar_sstl;
			newarp19->oar_tpln = arp19->oar_spln;
			np = (caddr_t) &newarp19[1];
			if (myipaddr (sc, np)) {
				nlen += 4;
				bcopy ((caddr_t) &p[lenoffs[0]], &np[4],
					lenoffs[3] - lenoffs[0]);
				nlen += lenoffs[3] - lenoffs[0];
				dst.sa_family = AF_UNSPEC;
				bcopy ((caddr_t) eptr->ether_shost,
					(caddr_t) &dst.sa_data[0], 6);
				bcopy ((caddr_t) &eptr->ether_type,
					(caddr_t) &dst.sa_data[12], 2);
				otto_simplesendpkt (sc, vc, msgbuf, nlen, &dst);
			}
			/* fall through */
		case INARP_REPLY:
			/* pick out source IP address */
			if (vc == 0 || vc >= OTTO_MAXVC) {
				printf("INARP_REPLY: bad vc %u\n", vc);
				panic ("otto: INARP_REPLY badvc");
			}
			if (arp19->oar_spln == 4 &&
			    (sc->data.vcs[vc].flags &
			      (OTTO_VC_PVC|OTTO_VC_DONT_ARP)) == OTTO_VC_PVC) {
				sc->data.vcs[vc].flags |=
					(OTTO_VC_ARPRCV|OTTO_VC_ARPRCV_RECENT);
				bzero ((char *) addr, sizeof (addr));
				bcopy ((caddr_t)p + lenoffs[2], (caddr_t)addr, 4);
				addr[1] = 0xffffffff;
				if (addr[0] != 0) {
					otto_map_set (&sc->data.ipvcmap,
						addr, vc);
				} else {
					printf ("otto_arp_rcv: empty type 19 ARP\n");
				}
			}
			break;

		case ARP_NAK:
			break;
		}
	} else if (htype == SARP) {
		if (len < sizeof (struct sarp)) {
			return (1);
		}
		lenoffs[0] = sizeof (struct sarp);
		lenoffs[1] = sarp->sar_hln;
		lenoffs[2] = sarp->sar_pln;
		lenoffs[3] = sarp->sar_hln;
		lenoffs[4] = sarp->sar_pln;
		sum = 0;
		for (i = 0; i != 5; i++) {
			sum += lenoffs[i];
			lenoffs[i] = sum;
		}
		if (lenoffs[4] > len) {
			return (1);
		}
		ptype = ntohs ((u_short) arp19->oar_pro);
		if (ptype != ANETTYPE_ETHER) {
			return (1);
		}
		if (sarp->sar_hln != 4 || sarp->sar_pln != 6) {
			return (1);
		}
		OTTO_STATS_INC (&sc->data, rcvsarp);
		bcopy (((caddr_t) p) + lenoffs[0], (caddr_t) &svc, 4);
		svc = ntohl ((unsigned) svc);
		op = ntohs ((u_short) sarp->sar_op);
		switch (op) {
		case ARPOP_REQUEST:
			if (svc == 0 || svc >= OTTO_MAXVC) {
				printf ("ARPOP_REQUEST: bad svc %u\n", svc);
			}
			if (svc != 0 && svc < OTTO_MAXVC &&
			    bcmp ((caddr_t) sc->is_addr,
			          ((caddr_t)p)+lenoffs[3], sarp->sar_pln)==0) {
				otto_arp_sendsarp (sc, svc, svc,
					eptr->ether_shost,
					((unsigned char*)p)+lenoffs[1], 1);
			}
			/* fall through */
		case ARPOP_REPLY:
			if (svc == 0 || svc >= OTTO_MAXVC) {
				printf ("ARPOP_REQUEST: bad svc %u\n", vc);
			}
			bzero ((char *) addr, sizeof (addr));
			bcopy ((caddr_t)p+lenoffs[1], (caddr_t)addr, 6);
			if ((addr[0] | addr[1]) != 0 && svc != 0 &&
					svc < OTTO_MAXVC) {
				otto_map_set (&sc->data.vcmap,addr,svc);
				bcopy ((caddr_t) addr, 
				    (caddr_t) sc->data.vcs[svc].addr, 6);
				sc->data.vcs[svc].flags |=
					OTTO_VC_UPDATED|OTTO_VC_RVC;
				sc->data.vcs[svc].flags &=
					~(OTTO_VC_STALE|OTTO_VC_PVC);
			}

			if (sc->data.arpstate.m != (struct mbuf *) 0 &&
			    bcmp ((caddr_t) sc->data.arpstate.uid,
				  (caddr_t) addr, 6)==0) {
				struct mbuf *m = sc->data.arpstate.m;
#ifdef AF_FWD
				dst.sa_family = AF_FWD;
#else
				dst.sa_family = AF_UNSPEC;
#endif
				bcopy ((caddr_t) sc->data.arpstate.uid,
					(caddr_t) &dst.sa_data[0], 6);
				bcopy ((caddr_t) sc->data.arpstate.srcuid,
					(caddr_t) &dst.sa_data[6], 6);
				bcopy ((caddr_t) &sc->data.arpstate.type,
					(caddr_t) &dst.sa_data[12], 2);
#ifdef __osf__
				ottooutput (&sc->is_if, m, &dst,
					(struct rtentry *) 0);
#else
				ottooutput (&sc->is_if, m, &dst);
#endif
				sc->data.arpstate.m = 0;
			}
			break;
		}
	} else {
		return (0);
	}
	return (1);
}

/* Send a type 19 arp on VC vc.
   Called with softc lock held. */
void otto_arp_sendarp19 (struct otto_softc *sc, unsigned vc)
{
	unsigned msgbuf[64];
	struct arp19 *newarp19 = (struct arp19 *) msgbuf;
	unsigned len = sizeof (struct arp19) + 8;
	caddr_t np;
	struct sockaddr dst;
	struct ether_header *eptr = (struct ether_header *) dst.sa_data;

	if (vc == 0 || vc >= OTTO_MAXVC) {
		printf ("otto_arp_sendarp19 badvc %u\n", vc);
		panic ("otto_arp_sendarp19: badvc");
	}

	sc->data.vcs[vc].flags |= OTTO_VC_ARPXMT;

	newarp19->oar_hrd = htons ((u_short) ARP19);
	newarp19->oar_pro = htons ((u_short) ETHERTYPE_IP);
	newarp19->oar_shtl = 0;
	newarp19->oar_sstl = ARP19_E164;
	newarp19->oar_op = htons ((u_short) INARP_REQUEST);
	newarp19->oar_spln = 4;
	newarp19->oar_thtl = 0;
	newarp19->oar_tstl = ARP19_E164;
	newarp19->oar_tpln = 4;

	np = (caddr_t) &newarp19[1];
	bzero (np, 8);
	if (myipaddr (sc, np)) {
		dst.sa_family = AF_UNSPEC;
		bcopy ((caddr_t)etherbroadcastaddr, (char*)eptr->ether_dhost,6);
		eptr->ether_type = ETHERTYPE_ARP;

		otto_simplesendpkt (sc, vc, msgbuf, len, &dst);
	}
}

#define RETRY_TIMEOUT	2	/* seconds between ARP retries */
#define ARP_REFRESH	30	/* seconds between refreshing ARP cache */
#define ARP_NRETRY	8	/* retries if no reply */
#define SHOULDARP(flags)	(((flags) & (OTTO_VC_PVC|OTTO_VC_DONT_ARP)) == \
					OTTO_VC_PVC)
		    /* flags indicate that this VC should be ARPed on */
#define SENTARP(flags)	((flags) & OTTO_VC_ARPXMT)
		    /* flags indicate that an ARP has been sent sometime */
#define RCVDARP(flags)	((flags) & OTTO_VC_ARPRCV)
		    /* flags indicate that an ARP has been received sometime */
#define RCVDRECENTARP(flags)	((flags) & OTTO_VC_ARPRCV_RECENT)
		    /* flags indicate that an ARP has been received sometime */

/* called once per second by watchdog routine.
   Sends type 9 ARPs on Classical IP PVCs, and expunges entries from
   the RVC ARP cache.
   Called with softc lock held */
void otto_arp_tick (struct otto_softc *sc)
{
	unsigned vc;
	int xvc;
	unsigned i;
	unsigned retry;
	struct otto_vcdata *vcs = sc->data.vcs;
	struct arpstate *as = &sc->data.arpstate;
	unsigned minarp = as->minarp;
	unsigned maxarp = as->maxarp;
	unsigned range = maxarp - minarp;
	unsigned n = (range+ARP_REFRESH-1) / ARP_REFRESH;
	struct otto_map_newmapping m;

	if (as->rearp > 0) {
		n = (range+9) / 10;
		as->rearp--;
	}

	vc = as->curvc;
	if (vc < minarp) {
		vc = minarp;
	}
	for (i = 0; i != n; i++) {
		if (vc >= maxarp) {
			vc = minarp;
		}
		if (vc == 0 || vc >= OTTO_MAXVC) {
			printf ("otto_arp_tick: bad VC %u\n", vc);
			panic ("otto_arp_tick: badvc");
		}
		if (SHOULDARP (vcs[vc].flags)) {
			vcs[vc].flags &= ~OTTO_VC_ARPRCV_RECENT;
			otto_arp_sendarp19 (sc, vc);
		}

		xvc = vc;
		/* Wait ARP_NRETRY retry timeout periods.  Retry on
		   first ARP_NRETRY-1 if there's no answer and we have
		   previously had an asnwer.  On the ARP_NRETRYth timeout,
		   don't retry, and wipe the VC from the cache.  */
		for (retry = 0; retry != ARP_NRETRY; retry++) {
			unsigned f;
			xvc -= RETRY_TIMEOUT*n;
			while (xvc < (int) minarp) {
				xvc += range;
			}
			if (xvc == 0 || xvc >= OTTO_MAXVC) {
				printf ("otto_arp_tick: bad xvc %d\n", xvc);
				panic ("otto_arp_tick badxvc");
			}
			f = vcs[xvc].flags;
			if (SHOULDARP (f) && SENTARP (f) &&
			    RCVDARP (f) && !RCVDRECENTARP (f)) {
				if (retry != ARP_NRETRY-1) {
					/* retry sending ARP */
					otto_arp_sendarp19 (sc, xvc);
				} else {
					/* wipe ARP cache */
					vcs[xvc].flags &= ~OTTO_VC_ARPRCV;
					while (otto_map_from_vc (
					    &sc->data.portvcmap, xvc, m.addr)) {
						otto_map_delete (
						   &sc->data.portvcmap, m.addr);
					}
					while (otto_map_from_vc (
					    &sc->data.ipvcmap, xvc, m.addr)) {
						otto_map_delete (
						    &sc->data.ipvcmap, m.addr);
					}
					while (otto_map_from_vc (
					    &sc->data.vcmap, xvc, m.addr)) {
						otto_map_delete (
						    &sc->data.vcmap, m.addr);
					}
				}
			}
		}
		vc++;
	}
	as->curvc = vc;

	if (as->newpvcs) {
		minarp = 0;
		maxarp = 0;

		for (vc = 1; vc != OTTO_MAXVC; vc++) {
			unsigned flags = vcs[vc].flags;
			if (SHOULDARP (flags)) {
				if (minarp == 0) {
					minarp = vc;
				}
				maxarp = vc+1;
				if (!SENTARP (flags)) {
					otto_arp_sendarp19 (sc, vc);
					vcs[vc].flags |= OTTO_VC_ARPRCV;
				}
			}
		}
		as->newpvcs = 0;
		if (minarp != maxarp && maxarp - minarp < 2*ARP_REFRESH) {
			if (maxarp > 2*ARP_REFRESH) {
				minarp = maxarp - 2*ARP_REFRESH;
			} else {
				maxarp = minarp + 2*ARP_REFRESH;
			}
		}
		as->minarp = minarp;
		as->maxarp = maxarp;
	}

	as->tick++;
	if ((as->tick & 3) == 0) {
		/* check freshness of ARP cache every 4 seconds */
		for (vc = 1; vc != OTTO_MAXVC; vc++) {
			unsigned flags = vcs[vc].flags;
			if ((flags & OTTO_VC_RVC) == 0) {
				continue;
			}
			if (flags & OTTO_VC_UPDATED) {
				vcs[vc].flags =
				   (flags & ~(OTTO_VC_USED|OTTO_VC_UPDATED));
			} else if (flags & OTTO_VC_USED) {
				if (flags & OTTO_VC_STALE) {
					bzero ((char *) m.addr,
						sizeof (m.addr));
					bcopy ((caddr_t) vcs[vc].addr,
						(caddr_t) m.addr, 6);
					if (m.addr[0] != 0 || 
					    m.addr[1] != 0) {
						otto_map_delete (
						    &sc->data.vcmap, m.addr);
						bzero ((caddr_t) vcs[vc].addr,
							6);
						OTTO_STATS_INC (&sc->data,
							zapcache);
					}
					vcs[vc].flags = (flags &
						~(OTTO_VC_USED|OTTO_VC_RVC));
				} else {
					vcs[vc].flags = (flags&~OTTO_VC_USED) |
						OTTO_VC_STALE;
				}
			}
		}

		/* and throw away old pkts waiting to be transmitted */
		if (sc->data.arpstate.m != (struct mbuf *) 0) {
			OTTOMFREEM (sc->data.arpstate.m);
			sc->data.arpstate.m = (struct mbuf *) 0;
		}
	}
	/* forget about arps we've done */
	sc->data.arpstate.ndone = 0;

	/* Decrement recent broadcast flag.  When it goes to zero,
	   we are allowed to broadcast one client packet again. */
	if (as->recentbcast != 0) {
		as->recentbcast--;
	}
}

/* The packet contained in "m" with ethernet type "type" is to be sent
   to UID "uid[0...5]" from UID "srcuid[0..5]" on interface "sc".
   otto_arp_resolve arranges to look up "uid" with a SARP request.
   If "vc" is non-zero, the SARP request should be sent on that vc.
   If it returns zero, the caller still owns the mbuf chain on return.
   If it returns non-zero, this call consumes the mbuf chain "m"---the
   caller should assume that it will be freed by otto_arp_resolve at
   some time in the future.
   Called with the otto softc lock not held.   Ths routine acquires and
   releases the lock internally. */
int otto_arp_resolve (struct otto_softc *sc, struct mbuf *m,
        unsigned char *uid, unsigned char *srcuid, unsigned short type, 
        unsigned vc)
{
	OTTO_SMP_LOCKTMP s;
	int rc = 0;

	OTTO_SMP_LOCK (&sc->data.lk_otto_softc, s);
		if (vc == 0) {
			int i;
			int n;
			/* If we haven't done this recently, we just
			   give the packet back to the caller for transmission.
			   Otherwise we save if for the ARP reply.
			   This limits the number of articificial broadcasts
			   to one every 3 seconds per interface. */
			if (sc->data.arpstate.recentbcast == 0) {
				sc->data.arpstate.recentbcast = 3;
			} else {
				if (sc->data.arpstate.m != (struct mbuf *) 0) {
					OTTOMFREEM (sc->data.arpstate.m);
					sc->data.arpstate.m = (struct mbuf *) 0;
				}
				rc = 1;
				sc->data.arpstate.m = m;
				bcopy ((caddr_t) uid,
					(caddr_t) sc->data.arpstate.uid, 6);
				bcopy ((caddr_t) srcuid,
					(caddr_t) sc->data.arpstate.srcuid, 6);
				sc->data.arpstate.type = type;
			}

			/* We limit the number of ARPs to any given UID
			   to be no greater than 2 per second.   We limit
			   the number of different UIDs ARPed in any second
			   to OTTO_MAXARP. */
			i = 0;
			n = sc->data.arpstate.ndone;
			while (i < n && i < OTTO_MAXARP && bcmp ((caddr_t) uid,
			     (caddr_t) sc->data.arpstate.done[i].uid, 6) != 0) {
				i++;
			}

			if (i < OTTO_MAXARP) {
				if (i == n) {
					bcopy ((caddr_t) uid, (caddr_t)
					   sc->data.arpstate.done[i].uid, 6);
					sc->data.arpstate.done[i].count = 0;
					sc->data.arpstate.ndone = i+1;
					vc = sc->data.snmp.pkt.arpVCIout;
				} else if (sc->data.arpstate.done[i].count==0) {
					sc->data.arpstate.done[i].count = 1;
					vc = sc->data.snmp.pkt.arpVCIout;
				}
			}
		}

		if (vc != 0) {
			otto_arp_sendsarp (sc, vc, 0, uid, uid, 0);
		}
	OTTO_SMP_UNLOCK (&sc->data.lk_otto_softc, s);
	return (rc);
}
