/* Code to perform AN2 flow control (FLOWmaster) credit resynchronization */

#include "otto_flow.h"
#include <sys/types.h>
#include <sys/time.h>
#ifdef __osf__
/* #include <sys/malloc.h>  ---can't be used in V1.3 */
#include <vm/vm_kern.h>
#else
#include <sys/kmalloc.h>
#endif
#include <sys/kernel.h>

#if 0
FlowMaster control messages are sent in both directions on a line 
using lineVCI=F, which circuit is never flow controlled itself. The 
value of F is obtained from the DEC-MIB. The FlowMaster mechanisms 
in the two directions on a line are operate independently, but use 
the same duplex control circuit. For a particular direction we have 
the line model: 

            >--------- cells ---------->
            
         U  <>---- control vci=F -----<>  D
         
            <--------- acks -----------<

where U stands for the upstream end (source of data cells) and D 
stands for the down stream end (source of acknowledgements).

Each Flowmaster control message is carried in an AAL5 packet. The 
layout of the data octets of the AAL5 packet is: 

      FM Control Message Format
     
      Meaning         Octets
      -------         ------

      Command       4  3  2  1    /* operation to be performed */

   FirstLineVCI     8  7  6  5    /* start of vci range */

    LastLineVCI    12 11 10  9    /* end of vci range */

       Token       16 15 14 13    /* matches responses to requests */

Note that all values are little-endian on the wire.

The protocol depends on three global constants:

   T       timeout in milliseconds
              (approximately 1 second?)
   E       max roundtrip delay for a line
              (50 milliseconds is enough)
   R       max number of outstanding requests allowed at one time
              (is 16 enough?)

There are two control message so far.

  SYNC: U -> D

    Command = 0x0F0000FF

    This command is used by U at line startup to cause D to start 
    returning FlowMaster acks to U; it is used thereafter to resynchronize 
    the credits for circuits on the line as needed. Once FlowMaster 
    acks are started in a direction on the line they cannot be turned 
    off except by dropping SONET carrier on the line, i.e. if U drops 
    carrier then D will notice and reinitialize all incoming vci's 
    to the state where they do not generate FlowMaster acks. 

    Before sending a SYNC message, U must arrange to not send data 
    cells on the FlowMaster vci's in the range [FirstLineVCI ... 
    LastLineVCI] until either 1) the period T+E passes or 2) a SACK 
    message with a matching Token value is received from D. U should 
    not reuse Token values in SYNC messages within the period T+E. 
    The FlowMaster vci's from U to D are any legal vci's on the line 
    that are not currently being used for CBR circuits and are not 
    permanent non-flow-controlled vci's such as vci=5, vci=16, and 
    vci=F. 
    
    Upon receiving a SYNC message D begins to observe the count of 
    buffered incoming cells for the FlowMaster vci's included in 
    the range [FirstLineVCI ... LastLineVCI]. Within the period T, 
    once D observes each of the FlowMaster vci's in the range with 
    no buffered cells and all of the acks generated by forwarding 
    cells for those vci's have had a chance to clear any pipelines 
    within D, D transmits a SACK with the matching token to U. If 
    the period T passes without seeing each of the indicated FlowMaster 
    vci's with empty cell buffers then D does not send a matching 
    SACK.

    D uses the arrival of SYNC requests as the trigger to start sending 
    acks back to U. The rules are: 

      1) D MAY NOT generate acks for ANY vci's on a line before the 
      first SYNC is received from U following line startup. 

      2) D must always be generating acks for all FlowMaster controlled 
      vci's that have been in the range of any SACKed SYNC requests 
      that have been received since line startup. 

      3) Once any SYNC message has been received following line startup 
      then D may generate acks for ANY vci's on the line.

    Note that these rules can be met by turning on acks for all vci's 
    after the first SYNC request is received.

    D is expected to be able to handle R outstanding SYNC requests 
    at one time. However, concurrent requests may not refer to overlapping 
    ranges of lineVCI values. Any arriving request that overlaps 
    with an outstanding request may be discarded. 
  

  SACK: D -> U

    Command = 0x0F0011FF

    When U receives a SACK message that matches the token of an outstanding 
    SYNC message it sets the maximum credit of the FlowMaster lineVCIs 
    in the associated range and then allows transmission of data 
    cells (again) on those that are active. The FirstLineVCI and 
    LastLineVCI fields of the SACK message are ignored by U. 

  SYNCBOTH: U -> D

    Command = 0x0F0022FF

    This message is used in exactly the same way as SYNC, except that it
    also triggers credit resynchronization in the reverse direction.

    Upon receiving SYNCBOTH, D should behave as though it received SYNC,
    and in addition it should send a SYNC whose VCI range
    FirstLineVCI,...,LastLineVCI covers at least the outgoing flow controlled
    VCs covered by the range of the SYNCBOTH message.  If there are no
    outgoing flow controlled VCs in the range, D need not send a SYNC
    message in response to the SYNCBOTH.

    The intention is that an end-point should use SYNCBOTH when a
    flow controlled virtual circuit has just been set up, and the peer
    end-point may or may not know about the virtual circuit yet.  This
    ensures that the credit balance on the virtual circuit will be
    initialized correctly and promptly at boths ends, regardless of
    which end is initialized first.
#endif

typedef unsigned fc_uint32;	/* a 32 bit unsigned integer */


/* The handling of time */

typedef struct timeval fc_time_t; /* a time */
/* The "struct timeval" is not ideal here.  It would be much better to
   have a counter that was incremented every clock interrupt. */

/* After the call, "*t" hold the current time plus at least "extrams"
   milliseconds. */
static void fc_now (fc_time_t *t, unsigned extrams)
{
	extern struct timeval time;
#ifdef __osf__
#ifdef TIME_READ
	TIME_READ (*t);
#else
	TIME_READ_LOCK ();
		*t = time;
	TIME_READ_UNLOCK ();
#endif
#else
	int s = splclock();
		*t = time;
	splx (s);
#endif
	t->tv_usec += extrams * 1000;
	while (t->tv_usec >= 1000000) {
		t->tv_usec -= 1000000;
		t->tv_sec++;
	}
}

/* fc_time_lt (fc_time_t *t1, fc_time_t *t2) returns non-zero iff
   time "*t1" is strictly before time "*t2" */
#define fc_time_lt(t1, t2) timercmp ((t1), (t2), <)

/* fc_time_gt (fc_time_t *t1, fc_time_t *t2) returns non-zero iff
   time "*t1" is strictly after time "*t2" */
#define fc_time_gt(t1, t2) timercmp ((t1), (t2), >)
/* end the handling of time */



struct fc_flowmasterpkt {	/* the format of a flowmaster packet */
			/* all fields are little-endian byte order*/
	fc_uint32 cmd;	/* operation ---see below */
#define FC_SYNC	0x0f0000ff /* Request to sync credits and enable flow control*/
#define FC_SACK 0x0f0011ff /* Response to FC_SYNC */
#define FC_SYNCBOTH 0x0f0022ff /* Like FC_SYNC, but requests that peer also
				  send an FC_SYNC message */

	fc_uint32 firstvc;	/* first VC in range */
	fc_uint32 lastvc;	/* last VC in range */
	fc_uint32 token;	/* a nonce to match replies and requests */
};


#define FC_TIMEOUT 1000	/* milliseconds -- time to wait for a response */
#define FC_MAX_RRT 200	/* milliseconds -- maximum round-trip time */
#define FC_MAXREQI 16	/* maximum number of outstanding incoming requests */
#define FC_MAXREQO 16	/* maximum number of outstanding outgoing requests */
#define FC_POLLTIME 200	/* time between examining VCs */
#define FC_MAXRETRY 4	/* maximum number of times we retry a request */

struct fc_ireq_state {	/* state for an outstanding incoming request */
	struct fc_ireq_state *next;	/* next in list */
	fc_time_t timeout;		/* when request times out */
	fc_time_t oktime;		/* when incoming request is satisfied */
	unsigned firstvc;		/* first vc in request packet */
	unsigned lastvc;		/* last vc in request packet */
	fc_uint32 token;		/* token in request packet */
	struct fc_ackstate as;		/* ackstate for driver in mstowait */
};
struct fc_oreq_state {	/* state for an outstanding outgoing request */
	struct fc_oreq_state *next;	/* next in list */
	fc_time_t timeout;		/* when request times out */
	unsigned retries;		/* times this has been retried */
	unsigned doboth;		/* => use a SYNCBOTH message */
	unsigned firstvc;		/* first vc in request packet */
	unsigned lastvc;		/* last vc in request packet */
	fc_uint32 token;		/* token in request packet */
};


/* There is one fc_state record per controller that does flow control */ 
struct fc_state {
	unsigned on;			/* link supports flow control */

	fc_uint32 token;		/* next token to use */
	unsigned doall;			/* resync all VCs when all outstanding
					   outgoing requests are done.

					   DOALL invariant: Whenever doall
					   is set to a non-zero value,
					   it must be the case that the
					   usedreqout list (below) is
					   non-empty. */
	unsigned doboth;		/* => use a SYNCBOTH message next time
					   all credits are resynced */
   	fc_time_t lasttime;		/* used to detect when time goes
					   backwards */

	struct fc_ireq_state reqin[FC_MAXREQI];	/* incomping requests */
	struct fc_oreq_state reqout[FC_MAXREQO]; /* outgoing requests */
	struct fc_ireq_state *freereqin;	/* free list for incoming reqs */
	struct fc_oreq_state *freereqout;	/* freelist for outgoing reqs */
	struct fc_ireq_state *usedreqin;	/* used list for incoming reqs */
					  /* in timeout order, earliest first */
	struct fc_oreq_state *usedreqout;	/* used list for outgoing reqs */
					  /* in timeout order, earliest first */

	struct fc_controller_state fccs;  /* controller state */
};


/* convert x to/from fc network byte order (little endian) */
#if defined(__alpha) || defined(MIPSEL)
#define fc_byteorder(_x) (_x)
#define fc_unbyteorder(_x) (_x)
#else
static fc_uint32 fc_byteorder (unsigned x)
{
	union { fc_uint32 a; unsigned char b[4]; } tmp;
	tmp.b[0] = x;		tmp.b[1] = x >> 8;
	tmp.b[2] = x >> 16;	tmp.b[3] = x >> 24;
	return (tmp.a);
}
static unsigned fc_unbyteorder (fc_uint32 x)
{
	union { fc_uint32 a; unsigned char b[4]; } tmp;
	tmp.a = x;
	return (((unsigned) tmp.b[0]) | (((unsigned) tmp.b[1]) << 8) |
		(((unsigned) tmp.b[2]) << 16) | (((unsigned) tmp.b[3]) << 24));
}
#endif


/* Initialize the flow control resync state machine and return a pointer to
   a struct fc_state.  Otherwise return (struct fc_state *) 0.
   Called with fcs lock held. */
struct fc_state *fc_init (struct fc_controller_state *fccs)
{
	struct fc_state *fcs;
#ifdef __osf__
	fcs = (struct fc_state *) kmem_alloc (kernel_map, sizeof (*fcs));
	/*
		MALLOC (fcs, struct fc_state *, sizeof (*fcs), M_DEVBUF, 0);
		This macro exists only in V2.0 and later.
	 */
#else
	KM_ALLOC (fcs, struct fc_state *, sizeof (*fcs), KM_DEVBUF, 0);
#endif
	if (fcs != (struct fc_state *) 0) {
		fcs->token = 0;
		fcs->fccs = *fccs;
		fc_now (&fcs->lasttime, 0);
		fc_restart (fcs);
	}
	return (fcs);
}

/* Reset the flow control state machine to the same state that it was in
   just after fc_init() returned "fcs", except that old tokens are not reused
   in new packets.
   This operation is intended to be used when the link loses carrier.
   It causes the fc module to believe that flow control is disabled on the link.
   Called with fcs lock held. */
void fc_restart (struct fc_state *fcs)
{
	unsigned i;
	fcs->on = 0;
	fcs->doall = 0;
	fcs->doboth = 0;
	/* initialize free and used lists */
	fcs->freereqin = (struct fc_ireq_state *) 0;
	fcs->freereqout = (struct fc_oreq_state *) 0;
	fcs->usedreqin = (struct fc_ireq_state *) 0;
	fcs->usedreqout = (struct fc_oreq_state *) 0;
	for (i = 0; i != FC_MAXREQI; i++) {
		fcs->reqin[i].next = fcs->freereqin;
		fcs->freereqin = &fcs->reqin[i];
	}
	for (i = 0; i != FC_MAXREQO; i++) {
		fcs->reqout[i].next = fcs->freereqout;
		fcs->freereqout = &fcs->reqout[i];
	}
}

/* insert outgoing request "*req" onto list "list" so that the list is
   in timeout order, with the shortest timeout first. */
static void fc_o_insert (struct fc_oreq_state **list, struct fc_oreq_state *req)
{
	struct fc_oreq_state *entry;
	while ((entry = *list) != (struct fc_oreq_state *) 0 &&
	       fc_time_gt (&req->timeout, &entry->timeout)) {
	       list = &entry->next;
	}
	req->next = entry;
	*list = req;
}

/* insert incoming request "*req" onto list "list" so that the list is
   in timeout order, with the shortest timeout first. */
static void fc_i_insert (struct fc_ireq_state **list, struct fc_ireq_state *req)
{
	struct fc_ireq_state *entry;
	while ((entry = *list) != (struct fc_ireq_state *) 0 &&
	       fc_time_gt (&req->timeout, &entry->timeout)) {
	       list = &entry->next;
	}
	req->next = entry;
	*list = req;
}

/* Generate a packet with fields "cmd", "firstvc", "lastvc", and "token"
   set according to the arguments.
   Called with the fcs lock held. */
static void fc_send (struct fc_state *fcs, fc_uint32 cmd,
	unsigned firstvc, unsigned lastvc, fc_uint32 token)
{
	struct fc_flowmasterpkt pkt;
	pkt.cmd = fc_byteorder (cmd);
	pkt.firstvc = fc_byteorder (firstvc);
	pkt.lastvc = fc_byteorder (lastvc);
	pkt.token = fc_byteorder (token);
	(*fcs->fccs.send) (fcs->fccs.arg, (unsigned char *) &pkt, sizeof (pkt));
}


/* Send a flow control resync request and queue an outstanding outgoing
   request structure for the VC range [first, last] if first <= last,
   and there is no overlapping outsatnding request.   Return non-zero
   iff there were no outstanding incoming or outgoing requests before
   the call, and there is such a request after the call.  If both is non-zero,
   the peero endpoint is requested to resync the VCs in the other
   direction too.   Called with fcs lock held. */
static int fc_do_resync (struct fc_state *fcs, unsigned first, unsigned last,
	unsigned both)
{
	struct fc_oreq_state *req;
	int emptyqs;

	if (first > last) {
		return (0);
	}
	if (fcs->doall) {
		/* if doall is set, the circuits will get resynced
		   by the timeout routine */
		fcs->doboth |= both;
		return (0);
	}

	/* check for overlap with outstanding request */ 
	req = fcs->usedreqout;
	while (req != (struct fc_oreq_state *) 0) {
		if (last >= req->firstvc && first <= req->lastvc) {
			/* overlap */
			if (first < req->firstvc || last > req->lastvc) {
				fcs->doall = 1;	/*see DOALL invariant above*/
				fcs->doboth |= both;
			}
			return (0);
		}
		req = req->next;
	}

	/* check that there's a free request structure */
	req = fcs->freereqout;
	if (req == (struct fc_oreq_state *) 0) {
		fcs->doall = 1;			/*see DOALL invariant above */
		fcs->doboth |= both;
		return (0);
	}

	emptyqs = (fcs->usedreqout == (struct fc_oreq_state *) 0 &&
	          fcs->usedreqin == (struct fc_ireq_state *) 0);
	/* set up request structure and send packet */
	fcs->freereqout = req->next;
	req->doboth = both;
	req->retries = 0;
	fc_now (&req->timeout, FC_TIMEOUT+FC_MAX_RRT); /* read clock */
	req->token = fcs->token;
	fcs->token++;
	req->firstvc = first;
	req->lastvc = last;
	fc_o_insert (&fcs->usedreqout, req);
	(*fcs->fccs.stopcircuits) (fcs->fccs.arg, first, last);
	fc_send (fcs, both? FC_SYNCBOTH: FC_SYNC, first, last, req->token);
	return (emptyqs);
}


/* This routine is called whenever an timeout might have occurred.
   It checks whether a timeout has indeed occurred, and if
   so, steps the state machine on. If not, it computes the next
   possible timeout, and arranges to be called at that time. 

   In the rare event that time goes backwards, all the outstanding
   incoming and outgoing requests are thrown away, and a single
   outgoing request to resynchronize all VCs is issued.

   TIMEOUT invariant: A timeout on fc_timeout is outstanding whenever
   at least one of fcs->usedreqout and fcs->usedreqin is non-empty.

   Called with the fcs lock not held. */
static void fc_timeout (struct fc_state *fcs)
{
	fc_locktmp_t s;
	unsigned callagain = 0;	/* non-zero => need to call fc_timeout again*/
	struct fc_ireq_state *ireq;
	struct fc_oreq_state *oreq;
	struct fc_ireq_state **pireq;
	struct fc_oreq_state **poreq;
	fc_time_t mytime;
	unsigned normal = 1;

	fc_now (&mytime, 0);		/* read clock */

	s = (*fcs->fccs.lock) (fcs->fccs.arg);
		if (!fcs->on) {
			(*fcs->fccs.startcircuits) (fcs->fccs.arg,
				1, FC_LASTVC, 0);
			(*fcs->fccs.unlock) (fcs->fccs.arg, s);
			return;
		}

		if (fc_time_lt (&mytime, &fcs->lasttime)) {
			normal = 0; /* time went backwards */
			fcs->doall = 1; /* see DOALL invariant above */
				 /* Setting doall here preserves the invariant
				    even though the usedreqout can become
				    empty.  This is because the code below
				    makes the list non-empty if doall is set */
			fcs->doboth = 1;
		}
		fcs->lasttime = mytime;

		/* check for expiry of outgoing requests */
		poreq = &fcs->usedreqout;
		while ((oreq = *poreq) != (struct fc_oreq_state *) 0) {
			if (normal && fc_time_lt (&mytime, &oreq->timeout)) {
				poreq = &oreq->next;	/* still waiting */
			} else if (oreq->retries >= FC_MAXRETRY || fcs->doall) {
				/* entry timed out */
				/* We've retried too much, or doall is set.
				   Restart the circuit, dequeue the request,
				   and cause all VCs to be resynced sometime
				   in the future.
				   If the request was to resync all VCs,
				   we take the link down temporarily to 
				   clear the fifo at the switch */
				if (oreq->firstvc == 1 &&
				    oreq->lastvc == FC_LASTVC) {
					(*fcs->fccs.zaplink) (fcs->fccs.arg);
				}
                                fcs->doall = 1; /* see DOALL invariant above. */
                                 /* Setting doall here preserves the invariant
                                    even though the usedreqout can become
                                    empty.  This is because the code below
                                    makes the list non-empty if doall is set */
                                fcs->doboth |= oreq->doboth;
                                (*fcs->fccs.startcircuits) (fcs->fccs.arg,
                                        oreq->firstvc, oreq->lastvc, 0);
                                *poreq = oreq->next;
                                oreq->next = fcs->freereqout;
                                fcs->freereqout = oreq;
			} else {
				/* entry timed out - retry */
				oreq->token = fcs->token;
				fcs->token++;
				oreq->retries++;
				fc_now (&oreq->timeout,
					oreq->retries*FC_TIMEOUT+FC_MAX_RRT);
				fc_send (fcs,
					oreq->doboth? FC_SYNCBOTH: FC_SYNC,
					oreq->firstvc, oreq->lastvc,
					oreq->token);
				poreq = &oreq->next;
			}
		}

		/* check for expiry of incoming requests */
		pireq = &fcs->usedreqin;
		while ((ireq = *pireq) != (struct fc_ireq_state *) 0) {
			if (normal && fc_time_lt (&mytime, &ireq->timeout)) {
				/* entry needs testing */
				if (fc_time_gt (&mytime, &ireq->oktime) ||
				    (*fcs->fccs.mstowait) (fcs->fccs.arg,
						ireq->firstvc, ireq->lastvc, 0,
						&ireq->as) == 0) {
					fc_send (fcs, FC_SACK, ireq->firstvc,
						ireq->lastvc, ireq->token);
					*pireq = ireq->next;
					ireq->next = fcs->freereqin;
					fcs->freereqin = ireq;
				} else {
					pireq = &ireq->next;
				}
			} else { /* entry timed out */
				*pireq = ireq->next;
				ireq->next = fcs->freereqin;
				fcs->freereqin = ireq;
			}
		}

		if ((oreq = fcs->usedreqout) != (struct fc_oreq_state *) 0) {
			callagain = 1; /* there's an outgoing request */
		} else if (fcs->doall) {
			unsigned doboth = fcs->doboth;
			/* no outstanding requests, except one for everything*/
			fcs->doall = 0;	/* see DOALL invariant above. */
			fcs->doboth = 0;
			(void) fc_do_resync (fcs, 1, FC_LASTVC, doboth);
			callagain = 1;
		}
		if ((ireq = fcs->usedreqin) != (struct fc_ireq_state *) 0) {
			callagain = 1; /* there's an incoming request */
		}
		if (callagain) { /* see TIMEOUT invariant above */
			timeout (fc_timeout, fcs, (hz*FC_POLLTIME)/1000);
		}
	(*fcs->fccs.unlock) (fcs->fccs.arg, s);
}

/* Send a flow control resync request for the VC range [first,last] 
   Called with fcs lock held. */
void fc_resync (struct fc_state *fcs, unsigned first, unsigned last,
	unsigned both)
{
	if (fcs->on && fc_do_resync (fcs, first, last, both)) {
		timeout (fc_timeout, fcs, (hz*FC_POLLTIME)/1000);
			/* see TIMEOUT invariant above */
	}
}

/* A SACK packet with token "token" has been received.  If its token
   matches an outstanding outgoing request, fc_rcv_sack removes it from
   the list of outstanding requests, and restarts the circuits in 
   the range.  Called with fcs lock held. */
static void fc_rcv_sack (struct fc_state *fcs, fc_uint32 token)
{
	struct fc_oreq_state *req;
	struct fc_oreq_state **preq;

	/* check for outstanding requests with a matching token */
	preq = &fcs->usedreqout;
	while ((req = *preq) != (struct fc_oreq_state *) 0) {
		if (token == req->token) { /* found -- start the circuits */
			(*fcs->fccs.startcircuits) (fcs->fccs.arg,
				req->firstvc, req->lastvc, 1);
			*preq = req->next;
			req->next = fcs->freereqout;
			fcs->freereqout = req;
			break;
		} else {
			preq = &req->next;
		}
	}
}


/* A SYNC packet "p" has been received.   
   fc_rcv_sync checks that the packet is legal, and that not too many
   outstanding requests exist.  If these conditions are satisfied,
   the driver is called to check when outstanding ACKs on the VCs will
   have drained.  If no ACKs are outsatnding, a response packet is generated,
   otherwise the request is added to the list of outsatnding incoming
   requests.
   Called with fcs lock held. */
static void fc_rcv_sync (struct fc_state *fcs, struct fc_flowmasterpkt *p)
{
	struct fc_ireq_state *req;
	unsigned ms;
	struct fc_ackstate as;
	unsigned emptyqs;

	/* check for overlap with outstanding request,
	   and for duplicate tokens */ 
	req = fcs->usedreqin;
	while (req != (struct fc_ireq_state *) 0) {
		if (p->lastvc >= req->firstvc && p->firstvc <= req->lastvc) {
			return;		/* overlap */
		}
		if (p->token == req->token) {
			return;		/* duplicate token */
		}
		req = req->next;
	}

	/* this is a legal request */
	ms = (*fcs->fccs.mstowait) (fcs->fccs.arg, p->firstvc,
		p->lastvc, 1, &as);
	if (ms == 0) {
		fc_send (fcs, FC_SACK, p->firstvc, p->lastvc, p->token);
	} else if ((req = fcs->freereqin) != (struct fc_ireq_state *) 0) {
		fcs->freereqin = req->next;	/* we a allocated record */
		if (ms > FC_TIMEOUT) {	/* so time won't overflow */ 
			ms = 2*FC_TIMEOUT;
		}
		fc_now (&req->timeout, FC_TIMEOUT);
		fc_now (&req->oktime, ms);
		req->firstvc = p->firstvc;
		req->lastvc = p->lastvc;
		req->token = p->token;
		req->as = as;
		emptyqs = (fcs->usedreqout == (struct fc_oreq_state *) 0 &&
			  fcs->usedreqin == (struct fc_ireq_state *) 0);
		fc_i_insert (&fcs->usedreqin, req);
		if (emptyqs) { /* see TIMEOUT invariant below */
			if (ms > FC_POLLTIME) {
				ms = FC_POLLTIME;
			}
			timeout (fc_timeout, fcs, (hz*ms)/1000);
		}
	}
}


/* This routine is called each time a flow control resynchronization
   packet is received.   
   Requires the packet in "*pkt" to have length "len" bytes.
   Called with the fcs lock held.  */
void fc_receive (struct fc_state *fcs, unsigned char *pkt, unsigned len)
{
	if (fcs->on && len >= sizeof (struct fc_flowmasterpkt)) {
		struct fc_flowmasterpkt p;
		bcopy ((char *) pkt, (char *) &p,
			sizeof (struct fc_flowmasterpkt));
		p.cmd = fc_unbyteorder (p.cmd);
		p.firstvc = fc_unbyteorder (p.firstvc);
		p.lastvc = fc_unbyteorder (p.lastvc);
		p.token = fc_unbyteorder (p.token);
		if (p.cmd == FC_SYNC) {
			fc_rcv_sync (fcs, &p);
		} else if (p.cmd == FC_SACK) {
			fc_rcv_sack (fcs, p.token);
		} else if (p.cmd == FC_SYNCBOTH) {
			fc_rcv_sync (fcs, &p);
			fc_resync (fcs, p.firstvc, p.lastvc, 0);
		}
	}
}

/* "fc_enable" is used by the driver to indicate that the link 
   currently supports flow control.    
   The "fc" module will not send resync messages on links that does not
   support flow control.
   Called with the fcs lock held.  See the section on LOCKING above. */
void fc_enable (struct fc_state *fcs)
{
	fcs->on = 1;
	fc_resync (fcs, 1, FC_LASTVC, 1);
}
