#include "otto_platform.h"
#include "otto_map.h"

#define OTTO_MAP_ADDRASSIGN(a,b) \
	( (a)[0] = (b)[0], (a)[1] = (b)[1], (a)[2] = (b)[2], \
	  (a)[3] = (b)[3], (a)[4] = (b)[4], (a)[5] = (b)[5] )
#define OTTO_MAP_ADDREQ(a,b) \
	( (a)[0] == (b)[0] && (a)[1] == (b)[1] && (a)[2] == (b)[2] && \
	  (a)[3] == (b)[3] && (a)[4] == (b)[4] && (a)[5] == (b)[5] )
#define OTTO_MAP_ADDRMASKCMP(a,b) \
	( ((a)[0] & (b)[1]) == (b)[0] && \
	  ((a)[1] & (b)[3]) == (b)[2] && \
	  ((a)[2] & (b)[5]) == (b)[4] )


struct otto_map_entry {
	struct otto_map_entry *next;
	struct otto_map_newmapping mapping;
};


/* Initializes an "struct otto_map" to unlocked and empty. */
void otto_map_init (struct otto_map *map)
{
	OTTO_SMP_LOCKINIT (&map->lock);
	map->list = (struct otto_map_entry *) 0;
	map->free = (struct otto_map_entry *) 0;
	map->m = (struct mbuf *) 0;
}

/* Search "map" for an entry that has address "addr[0..OTTO_MAP_NADDR-1]".
   If none is found, create one.
   In any case, set the entry's address to "addr[0..OTTO_MAP_NADDR-1]", 
   and its vc to "vc". */
void otto_map_set (struct otto_map *map, unsigned addr[], unsigned vc)
{
	OTTO_SMP_LOCKTMP s;
	struct otto_map_entry *p;

	OTTO_SMP_LOCK (&map->lock, s);
again:
	p = map->list;
	while (p != (struct otto_map_entry *) 0) {
		if (OTTO_MAP_ADDREQ (addr, p->mapping.addr)) {
			break;
		}
		p = p->next;
	}
	if (p == (struct otto_map_entry *) 0) {
		p = map->free;
		if (p == (struct otto_map_entry *) 0) {
			struct mbuf *m;
			int i;
			OTTO_SMP_UNLOCK (&map->lock, s);
			OTTOMGET(m);
			OTTO_SMP_LOCK (&map->lock, s);
			i = m->m_len;
			m->m_next = map->m;
			map->m = m;
			p = mtod (m, struct otto_map_entry *);
			while (i >= sizeof (*p)) {
				p->next = map->free;
				map->free = p;
				p++;
				i -= sizeof (*p);
			}
			goto again;
		} else {
			OTTO_MAP_ADDRASSIGN (p->mapping.addr, addr);
			map->free = p->next;
			p->next = map->list;
			map->list = p;
		}
	}
	p->mapping.vc = vc;
	OTTO_SMP_UNLOCK (&map->lock, s);
}

/* Search "map" for an entry that has address "addr[0..OTTO_MAP_NADDR-1]".
   If one is found, delete it. */
void otto_map_delete (struct otto_map *map, unsigned addr[])
{
	OTTO_SMP_LOCKTMP s;
	struct otto_map_entry **pp;
	struct otto_map_entry *p;

	OTTO_SMP_LOCK (&map->lock, s);
	pp = &map->list;
	while ((p = *pp) != (struct otto_map_entry *) 0) {
		if (OTTO_MAP_ADDREQ (addr, p->mapping.addr)) {
			*pp = p->next;
			p->next = map->free;
			map->free = p;
			break;
		}
		pp = &p->next;
	}
	OTTO_SMP_UNLOCK (&map->lock, s);
}

/* Search "map" for an entry that has vc "vc".
   If such an entry is found, set "addr[0..OTTO_MAP_NADDR-1]" to the
   address in the entry, and return 1.  Otherwise return 0. */
int otto_map_from_vc (struct otto_map *map, unsigned vc, unsigned addr[])
{
	OTTO_SMP_LOCKTMP s;
	struct otto_map_entry **pp;
	struct otto_map_entry *p;
	int found = 0;

	OTTO_SMP_LOCK (&map->lock, s);
	pp = &map->list;
	while ((p = *pp) != (struct otto_map_entry *) 0) {
		if (vc == p->mapping.vc) {
			OTTO_MAP_ADDRASSIGN (addr, p->mapping.addr);
			*pp = p->next;
			p->next = map->list;
			map->list = p;
			found = 1;
			break;
		}
		pp = &p->next;
	}
	OTTO_SMP_UNLOCK (&map->lock, s);
	return (found);
}

/* Search "map" for an entry that has address "addr[0..OTTO_MAP_NADDR-1]".
   If such an entry is found, return the vc in the entry.  Otherwise return 0.*/
unsigned otto_map_to_vc (struct otto_map *map, unsigned addr[])
{
	OTTO_SMP_LOCKTMP s;
	struct otto_map_entry **pp;
	struct otto_map_entry *p;
	unsigned vc = 0;

	OTTO_SMP_LOCK (&map->lock, s);
	pp = &map->list;
	while ((p = *pp) != (struct otto_map_entry *) 0) {
		if (OTTO_MAP_ADDREQ (addr, p->mapping.addr)) {
			vc = p->mapping.vc;
			*pp = p->next;
			p->next = map->list;
			map->list = p;
			break;
		}
		pp = &p->next;
	}
	OTTO_SMP_UNLOCK (&map->lock, s);
	return (vc);
}

/* Search "map" for an entry such that
                "entry->addr[0] == (entry->addr[1] & addr[0]) &&
                 entry->addr[2] == (entry->addr[3] & addr[1]) &&
                 ...
                 entry->addr[OTTO_MAP_NADDR-2] ==
               	    (entry->addr[OTTO_MAP_NADDR-2] & addr[OTTO_MAP_NADDR/2-1])"
   If such an entry is found, return the vc in the entry.  Otherwise return 0.

   This procedure is intended to be used for mapping entire IP networks to VCs.
   */
unsigned otto_map_ip_to_vc (struct otto_map *map, unsigned addr[])
{
	OTTO_SMP_LOCKTMP s;
	struct otto_map_entry **pp;
	struct otto_map_entry *p;
	unsigned vc = 0;

	OTTO_SMP_LOCK (&map->lock, s);
	pp = &map->list;
	while ((p = *pp) != (struct otto_map_entry *) 0) {
		if (OTTO_MAP_ADDRMASKCMP (addr, p->mapping.addr)) {
			vc = p->mapping.vc;
			*pp = p->next;
			p->next = map->list;
			map->list = p;
			break;
		}
		pp = &p->next;
	}
	OTTO_SMP_UNLOCK (&map->lock, s);
	return (vc);
}


/* If there are N entries in the map "map", 
   Copies MIN(N, count) entries from "map"  into "mapping[0..MIN(count-1,N-1)]"
   and returns MAX(N, count) or -1 on error.
   If touser is true, the copy is done with copyout(). */
int otto_map_copy (struct otto_map *map,
        struct otto_map_newmapping mapping[], unsigned count,
        int touser)
{
	OTTO_SMP_LOCKTMP s;
	unsigned n = 0;
	struct otto_map_entry *p;
	int error = 0;

	OTTO_SMP_LOCK (&map->lock, s);
	p = map->list;
	while (p != (struct otto_map_entry *) 0 && n != count) {
		if (touser) {
			error = copyout ((caddr_t) &p->mapping,
				(caddr_t) &mapping[n], sizeof (p->mapping));
		} else {
			bcopy ((char *) &p->mapping, (char *) &mapping[n],
				sizeof (p->mapping));
		}
		n++;
		p = p->next;
	}
	OTTO_SMP_UNLOCK (&map->lock, s);
	return (error? -1: n);
}

/* For each entry in the map "map", "func" is called with arguments
   "arg" and "mapping", where "mapping" points to the mapping represented
   by the entry.    On any particular entry, if "func" returns 0, the 
   iteration proceeds to the next entry.   If it returns non-zero, 
   the iteration terminates.   If there is no next entry, the iteration
   terminates.  In either case, otto_map_iterate returns the rteturn value
   of the last call to func, or 0 if there were no calls.

   func may not block.
   It is called under the map's internal lock. */
void *otto_map_iterate (struct otto_map *map, otto_iterfunc func, void *arg)
{
	void *ret = 0;
	struct otto_map_entry *p;
	OTTO_SMP_LOCKTMP s;

	OTTO_SMP_LOCK (&map->lock, s);
	p = map->list;
	while (p != (struct otto_map_entry *) 0 && ret == 0) {
		ret = (*func) (arg, &p->mapping);
		p = p->next;
	}
	OTTO_SMP_UNLOCK (&map->lock, s);
	return (ret);
}
