#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <netdb.h>
#include "otto_ioctl.h"

static char *progname;

static struct {
	char *name;
	unsigned num;
} encaps[] = {
	{ "atmf", OTTO_EN_ATMF },
	{ "sle",  OTTO_EN_SLE },
	{ "fddi", OTTO_EN_FDDI },
	{ "raw",  OTTO_EN_RAW },
	{ "llc",  OTTO_EN_LLC },
	{ "cred", OTTO_EN_CRED },
	{ (char *) 0,  0 }
};


#define MAX_MAPPINGS 4096
#define SCHEDSIZE 4096

void usage ()
{
	fprintf (stderr, "usage: %s [-I if] -a [-n] [-A] [vc]\n", progname);
	fprintf (stderr, "usage: %s [-I if] -s pvc [-enc e] [-type t] [-fc|-nofc] [-bw mbs] [default|ip_addr|UID]\n", progname);
	fprintf (stderr, "usage: %s [-I if] -d [pvc|ip_address|ip_network|UID] ...\n", progname);
	fprintf (stderr, "usage: %s [-I if] -f file\n", progname);
	fprintf (stderr, "usage: %s [-I if] -x [file]\n", progname);
	exit (2);
}

/* this ioctl routine provides for backward compatibility with old
   drivers */
static int doioctl (int fd, unsigned cmd, void *data)
{
	static int old = 0;
	int rc;

	switch (cmd) {
	case OTTO_GET_ENTRY2:
		if (!old) {
			rc = ioctl (fd, cmd, data);
			if (rc == -1 && errno == EINVAL) {
				old = 1;
			}
		}
		if (old) {
			struct otto_map_mapping table[MAX_MAPPINGS];
			otto_table_t t;
			otto_newtable_t *nt = (otto_newtable_t *) data;

			cmd = OTTO_GET_ENTRY;
			t.ip = nt->ip;
			t.length = nt->length;
			t.table = table;
			rc = ioctl (fd, cmd, (char *) &t);
			if (rc == 0) {
				int i;
				nt->ip = t.ip;
				nt->length = t.length;
				for (i = 0; i != t.length; i++) {
					bzero ((char *) nt->table[i].addr,
						sizeof (nt->table[i].addr));
					nt->table[i].addr[0]=t.table[i].addr[0];
					nt->table[i].addr[1]=t.table[i].addr[1];
					nt->table[i].vc = t.table[i].vc;
				}
			}
		}
		return (rc);
	case OTTO_DEL_ENTRY2:
	case OTTO_INS_ENTRY2:
		if (!old) {
			rc = ioctl (fd, cmd, data);
			if (rc == -1 && errno == EINVAL) {
				old = 1;
			}
		}
		if (old) {
			otto_mapping_t m;
			otto_newmapping_t *nm = (otto_newmapping_t *) data;
			if (cmd == OTTO_DEL_ENTRY2) {
				cmd = OTTO_DEL_ENTRY;
			} else {
				cmd = OTTO_INS_ENTRY;
			}
			m.ip = nm->ip;
			m.mapping.addr[0] = nm->mapping.addr[0];
			m.mapping.addr[1] = nm->mapping.addr[1];
			m.mapping.vc = nm->mapping.vc;
			rc = ioctl (fd, cmd, (char *) &m);
		}
		return (rc);
	default:
		return (ioctl (fd, cmd, data));
	}
}

/* Parse a string into a 48-bit uid, put the result in m->mapping.addr,
   zero m.ip and m.mapping.vc, and return 1.   If the string cannot be parsed,
   return 0. */
int parseuid (char *str, otto_newmapping_t *m)
{
	unsigned char *uid;
	unsigned iuid[6];
	static char *fmts[] = { "%x-%x-%x-%x-%x-%x", "%x:%x:%x:%x:%x:%x" };
	int i;
	int j;

	for (j = 0; j != sizeof (fmts) / sizeof (fmts[0]); j++) {
		if (sscanf (str, fmts[j],
			    &iuid[0], &iuid[1], &iuid[2],
			    &iuid[3], &iuid[4], &iuid[5]) == 6) {
		    	bzero ((char *) m, sizeof (*m));
			uid = (unsigned char *) m->mapping.addr;
			for (i = 0; i != 6; i++) {
				uid[i] = iuid[i];
			}
			return (1);
		}
    	}
    	return (0);
}

/* Parse a string into an ip address with optional protocol and port number(s).
   Put the result in m->mapping.addr, set m.ip to 1, zero m.mapping.vc, and
   return 1.   If the string cannot be parsed, return 0. */
int parseip (char *str, otto_newmapping_t *m)
{
	char *sl = strchr (str, '/');
	unsigned iuid[4];
	unsigned char uid[8];
	char hoststring[1024];
	int i;
	int count;
	struct hostent *host;
	unsigned sport;
	unsigned dport;
	unsigned protocol;

	if (sl != (char *) 0) {
		i = 0;
		while (&str[i] != sl) {
			hoststring[i] = str[i];
			i++;
		}
		hoststring[i] = '\0';
		str = hoststring;
	}
	bzero ((char *) uid, sizeof (uid));
	count = sscanf (str, "%d.%d.%d.%d",
		&iuid[0], &iuid[1], &iuid[2], &iuid[3]);
	if (count > 1) {
		for (i = 0; i != count && i != 4; i++) {
			uid[i] = iuid[i];
			uid[i+4] = 0xff;
		}
	} else if ((host = gethostbyname (str)) != 0) {
		bcopy ((char *) host->h_addr, (char *) uid, 4);
		uid[4] = 0xff;
		uid[5] = 0xff;
		uid[6] = 0xff;
		uid[7] = 0xff;
	} else {
		return (0);
	}
	bzero ((char *) m, sizeof (*m));
	m->ip = 1;
	bcopy ((char *) uid, (char *) m->mapping.addr, 8);

	if (sl != (char *) 0) {
		protocol = 6;
		count = sscanf (sl+1, "tcp%u,%u", &dport, &sport);
		if (count == 0) {
			protocol = 17;
			count = sscanf (sl+1, "udp%u,%u", &dport, &sport);
		}
		if (count > 0) {
			m->mapping.addr[4] = protocol;
			m->mapping.addr[5] = 0xffffffff;
			m->mapping.addr[2] |= htons (dport) << 16;
			m->mapping.addr[3] |= 0xffff0000;
			if (count > 1) {
				m->mapping.addr[2] |= htons (sport);
				m->mapping.addr[3] |= 0xffff;
			}
		}
	}
	return (1);
}

static int cmp (struct otto_map_newmapping *a, struct otto_map_newmapping *b)
{
	if (a->vc > b->vc) {
		return (1);
	} else if (a->vc < b->vc) {
		return (-1);
	} else {
		return (0);
	}
}

static void printvc (unsigned vc, int fd, double bwconv, unsigned def,
     int nonames,
     struct otto_map_newmapping table0[],
     struct otto_map_newmapping table1[])
{
	otto_vcsetup_t vcsetup;
	struct in_addr in;
	struct hostent *host;
	char *name;
	int j;
	unsigned next0 = 0;
	unsigned next1 = 0;

	vcsetup.vc = vc;
	if (ioctl (fd, OTTO_GET_VC, (char *) &vcsetup) == -1) {
		perror ("OTTO_GET_VC");
		exit (2);
	}

	printf ("%4u ", vc);  

	j = 0;
	while (encaps[j].num != vcsetup.encap &&
	       encaps[j].name != (char *) 0) {
		j++;
	}

	if (vcsetup.useflowcontrol) {
		printf ("-fc   ");
	} else {
		printf ("-nofc ");
	}

	if (encaps[j].name == (char *) 0) {
		printf ("-enc %4d ", vcsetup.encap);
	} else {
		printf ("-enc %-4s ", encaps[j].name);
	}

	if (vcsetup.encap == OTTO_EN_RAW) {
		printf ("-type 0x%04x ", vcsetup.etype);
	}

	if (vcsetup.useschedule) {
		printf ("-bw %6.2f ", bwconv * (double) vcsetup.bandwidth);
	} else {
		printf ("-bw abr    ");
	}

	if (vcsetup.ignore) {
		printf ("-ign ");
	}

	if (vc == def) {
		printf ("default ");
	}

	while (table0[next0].vc == vc) {
		name = "";
		if (table0[next0].addr[1] == 0xffffffff &&
		    !nonames) {
			host = gethostbyaddr (&table0[next0].addr[0],
				4, AF_INET);
			if (host) {
				name = host->h_name;
			}
		}
		in.s_addr = table0[next0].addr[0];
		if (table0[next0].addr[3] != 0) {
			unsigned x = table0[next0].addr[2] &
					table0[next0].addr[3];
			char *p = "";
			if (table0[next0].addr[5] == 0xffffffff) {
				if (table0[next0].addr[4] == 17) {
					p = "udp";
				} else if (table0[next0].addr[4] == 6) {
					p = "tcp";
				}
			}
			printf ("%s/%s%u,%u ",
				inet_ntoa (in), p,
				ntohs ((x >> 16) & 0xffff),
				ntohs (x & 0xffff));
			if (name[0] != '\0') {
				printf ("%s/%s%u,%u ", name, p,
					ntohs ((x >> 16) & 0xffff),
					ntohs (x & 0xffff));
			}
		} else {
			printf ("%s %s ", inet_ntoa (in), name);
		}
		next0++;
	}

	while (table1[next1].vc == vc) {
		unsigned char *p = (unsigned char *) table1[next1].addr;
		printf ("%02x-%02x-%02x-%02x-%02x-%02x ",
		    p[0], p[1], p[2], p[3], p[4], p[5]);
		next1++;
	}

	if (vcsetup.bandwidth != 0 && !vcsetup.useschedule) {
		printf ("# %d slots in schedule",
			vcsetup.bandwidth);
	}
	printf ("\n");
}

void dump_mappings (int fd, int all, int nonames, int argc, char *argv[])
{
	otto_newtable_t t0;
	otto_newtable_t t1;
	unsigned i;
	struct otto_map_newmapping table0[MAX_MAPPINGS+1];
	struct otto_map_newmapping table1[MAX_MAPPINGS+1];
	unsigned def;
	otto_sched_t sched;
	unsigned sch[SCHEDSIZE];
	unsigned vcbw[2048];
	unsigned t0pos[2048];
	unsigned t1pos[2048];
	double bwconv;

	if (ioctl (fd, OTTO_GET_DEFVC, (char *) &def) == -1) {
		perror ("OTTO_GET_DEFVC");
		exit (2);
	}
	t0.ip = 1;
	t0.length = MAX_MAPPINGS;
	t0.table = table0;
	if (doioctl (fd, OTTO_GET_ENTRY2, (char *) &t0) == -1) {
		perror ("OTTO_GET_ENTRY2");
		exit (2);
    	}
	qsort (table0, t0.length, sizeof (table0[0]), cmp);
	table0[t0.length].vc = ~0;

	t1.ip = 0;
	t1.length = MAX_MAPPINGS;
	t1.table = table1;
	if (doioctl (fd, OTTO_GET_ENTRY2, (char *) &t1) == -1) {
		perror ("OTTO_GET_ENTRY2");
		exit (2);
	}
	qsort (table1, t1.length, sizeof (table1[0]), cmp);
	table1[t1.length].vc = ~0;

	sched.length = sizeof (sch) / sizeof (sch[0]);
	sched.table = sch;
	if (ioctl (fd, OTTO_GET_SCHED, (char *) &sched) == -1) {
		perror ("OTTO_GET_SCHED");
		exit (2);
	}
	bzero (vcbw, sizeof (vcbw));
	for (i = 0; i != sched.length; i++) {
		vcbw[sch[i]]++;
	}

	{
		unsigned next0 = 0;
		unsigned next1 = 0;
		for (i = 0; i != 2048; i++) {
			t0pos[i] = next0;
			while (table0[next0].vc == i) {
				next0++;
			}
			t1pos[i] = next1;
			while (table1[next1].vc == i) {
				next1++;
			}
		}
	}

	bwconv = 135.63 / (double) sched.length;
	if (argc > 0) {
		int argn;
		for (argn = 0; argn < argc; argn++) {
			i = atoi (argv[argn]);
			if (i < 1 || i >= 2048) {
				fprintf (stderr, "%s: bad vc\n", argv[argn]);
			} else {
				printvc (i, fd, bwconv, def, nonames,
					&table0[t0pos[i]], &table1[t1pos[i]]);
			}
		}
	} else {
		for (i = 1; i != 2048; i++) {
			if (all || vcbw[i] != 0 || i == def || 
			    table0[t0pos[i]].vc == i ||
			    table1[t1pos[i]].vc == i) {
				printvc (i, fd, bwconv, def, nonames,
					&table0[t0pos[i]], &table1[t1pos[i]]);
			}
		}
	}
}



void delete_mapping (int fd, char *a, unsigned def)
{
	otto_newmapping_t m;
	int remvc = 0;
	int deldef = 0;
	int delmap = 0;

	bzero ((char *) &m, sizeof (m));
	if (strcmp (a, "default") == 0) {
		deldef = 1;
	} else if (parseuid (a, &m)) {
		delmap = 1;
	} else if (parseip (a, &m)) {
		delmap = 1;
	} else if (isdigit (a[0])) {
		m.ip = 0;
		m.mapping.vc = strtoul (a, (char **) 0, 0);
		remvc = 1;
		delmap = 1;
		if (m.mapping.vc == def) {
			deldef = 1;
		}
	} else {
		fprintf (stderr, "%s: %s: unknown host/address or vc syntax\n",
			progname, a);
		exit (2);
	}
	if (deldef) {
		unsigned vc = 0;
		if (ioctl (fd, OTTO_SET_DEFVC, (char *) &vc) == -1) {
			perror ("OTTO_SET_DEFVC");
			exit (2);
		}
	}
	if (remvc) {
		otto_vcsetup_t vcsetup;

		vcsetup.vc = m.mapping.vc;
		if (ioctl (fd, OTTO_GET_VC, (char *) &vcsetup) == -1) {
			perror ("OTTO_GET_VC");
			exit (2);
		}
		vcsetup.vc = m.mapping.vc;
		vcsetup.encap = 0;
		vcsetup.etype = 0;
		vcsetup.useflowcontrol = 1;
		vcsetup.ignore = 1;
		vcsetup.useschedule = 0;
		vcsetup.bandwidth = 0;
		vcsetup.dontarp = 0;
		if (ioctl (fd, OTTO_SET_VC, (char *) &vcsetup) == -1) {
			perror ("OTTO_SET_VC");
			exit (2);
		}
	}
	if (delmap) {
		if (doioctl (fd, OTTO_DEL_ENTRY2, (char *) &m) == -1) {
			perror ("OTTO_DEL_ENTRY");
			exit (2);
		}
	}
}

void setvc (int fd, int argc, char **argv, int line, unsigned schedlen)
{
	int argn;
	unsigned vc = strtoul (argv[0], (char **) 0, 0);
	otto_vcsetup_t vcsetup;
	int i;
	int def = 0;
	otto_newmapping_t mappings[128];
	int n = 0;

	if (vc == 0) {
		fprintf (stderr, "%s: expected non-zero VC on line %d\n",
			progname, line);
		exit (2);
	}

	vcsetup.vc = vc;
	vcsetup.encap = -1;
	vcsetup.etype = -1;
	vcsetup.useflowcontrol = -1;
	vcsetup.ignore = 0;
	vcsetup.useschedule = -1;
	vcsetup.bandwidth = -1;
	vcsetup.dontarp = -1;

	for (argn = 1; argn != argc; argn++) {
		if (strcmp (argv[argn], "-enc") == 0 ||
		    strcmp (argv[argn], "-encap") == 0) {
			if (argn+1 == argc || vcsetup.encap != -1) {
				usage ();
			}
			argn++;
			for (i = 0; encaps[i].name != (char *) 0; i++) {
				if (strcmp (encaps[i].name, argv[argn]) == 0) {
					vcsetup.encap = encaps[i].num;
					break;
				}
			}
			if (encaps[i].name == (char *) 0) {
				if (isdigit (argv[argn][0])) {
					vcsetup.encap = strtoul (argv[argn],
						(char **) 0, 0);
				} else {
					fprintf (stderr,
					  "%s: line %d: known encapsulations:",
					  progname, line);

					for (i=0;encaps[i].name!=(char *)0;i++){
						fprintf (stderr, " %s",
							encaps[i].name);
					}
					fprintf (stderr, "\n");
					exit (2);
				}
			}
		} else if (strcmp (argv[argn], "-type") == 0) {
			if (argn+1 == argc || vcsetup.etype != -1) {
				usage ();
			}
			argn++;
			if (strcmp (argv[argn], "ip") == 0) {
				vcsetup.etype=0x0800;
			} else {
				vcsetup.etype=strtoul (argv[argn],(char **)0,0);
			}
		} else if (strcmp (argv[argn], "-fc") == 0) {
			if (vcsetup.useflowcontrol != -1) {
				usage ();
			}
			vcsetup.useflowcontrol = 1;
		} else if (strcmp (argv[argn], "-nofc") == 0) {
			if (vcsetup.useflowcontrol != -1) {
				usage ();
			}
			vcsetup.useflowcontrol = 0;
		} else if (strcmp (argv[argn], "-bw") == 0 ||
			   strcmp (argv[argn], "-cbr") == 0) {
			if (argn+1 == argc || vcsetup.bandwidth != -1) {
				usage ();
			}
			argn++;
			if (strcmp (argv[argn], "abr") == 0) {
				vcsetup.bandwidth = 0;
				vcsetup.useschedule = 0;
			} else {
				vcsetup.bandwidth = (((float) schedlen) *
				    atof (argv[argn]) / 135.63) + 0.5;
				vcsetup.useschedule = 1;
			}
		} else if (strcmp (argv[argn], "default") == 0) {
			vcsetup.dontarp = 1;
			def = 1;
		} else if (parseuid (argv[argn], &mappings[n])) {
			mappings[n].mapping.vc = vc;
			vcsetup.dontarp = 1;
			n++;
		} else if (parseip (argv[argn], &mappings[n])) {
			mappings[n].mapping.vc = vc;
			vcsetup.dontarp = 1;
			n++;
		} else {
			fprintf (stderr, "%s: %s: unknown host/address\n",
				progname, argv[argn]);
			exit (2);
		}
	}

	if (ioctl (fd, OTTO_SET_VC, (char *) &vcsetup) == -1) {
		perror ("OTTO_SET_VC");
		exit (2);
	}
	if (def) {
		if (ioctl (fd, OTTO_SET_DEFVC, (char *) &vc) == -1) {
			perror ("OTTO_SET_DEFVC");
			exit (2);
		}
	}

	for (i = 0; i != n; i++) {
		if (doioctl (fd, OTTO_INS_ENTRY2, (char *) &mappings[i]) == -1) {
			perror ("OTTO_INS_ENTRY2");
		    exit (2);
		}
	}
}

int main (int argc, char **argv)
{
	int fd;
	char *arg;
	int argn;
	char *interface = "otto0";
	char path[1024];

	progname = argv[0];
	if (strrchr (progname, '/') != (char *) 0) {
		progname = strrchr (progname, '/') + 1;
	}

	argn = 1;
	if (argn < argc) {
		if (argv[argn][0] == '-' && argv[argn][1] == 'I') {
			if (argv[argn][2] != '\0') {
				interface = &argv[argn][2];
			} else if (argn+1 < argc) {
				interface = argv[++argn];
			} else {
				usage ();
			}
			argn++;
		}
	}

	if (argn >= argc) {
		usage ();
	}
	arg = argv[argn++];

	strcpy (path, "/dev/");
	strcat (path, interface);

	if ((fd = open (path, O_RDONLY, 0)) < 0) {
		perror (path);
		exit (1);
	}
	if (arg[0] == '-' && strchr (arg, 'a') != (char *) 0) {
		int all = (strchr (arg, 'A') != (char *) 0);
		int nonames = (strchr (arg, 'n') != (char *) 0);
		while (argn < argc && argv[argn][0] == '-') {
			if (strchr (argv[argn], 'A') != (char *) 0) {
				all = 1;
			}
			if (strchr (argv[argn], 'n') != (char *) 0) {
				nonames = 1;
			}
			argn++;
		}
		dump_mappings (fd, all, nonames, argc-argn, &argv[argn]);
	} else if (strcmp (arg, "-f") == 0) {
		FILE *fp;
		unsigned char buf[1024];
		int line;
		otto_sched_t sched;

		if (argn+1 != argc) {
			usage ();
		}
		sched.length = 0;
		if (ioctl (fd, OTTO_GET_SCHED, (char *) &sched) == -1) {
			perror ("OTTO_GET_SCHED");
			exit (2);
		}
		if (strcmp (argv[argn], "-") == 0) {
			fp = stdin;
		} else {
			fp = fopen (argv[argn], "r");
		}
		if (fp == NULL) {
			perror (argv[argn]);
			exit (2);
		}
		line = 1;
		while (fgets (buf, sizeof (buf), fp) != NULL) {
			unsigned char *p = buf;
			int newargc = 0;
			char *newargv[128];

			for (;;) {
				while (isspace (*p) && *p != '\0') {
					p++;
				}
				if (*p == '#' || *p == '\0' || newargc == 127) {
					break;
				}
				newargv[newargc++] = (char *) p;
				while (!isspace (*p) && *p != '\0') {
					p++;
				}
				if (*p != '\0') {
					*p++ = '\0';
				}
			}
			newargv[newargc] = (char *) 0;
			if (newargc > 0) {
				setvc (fd, newargc, newargv, line++,
					sched.length);
			}
		}
		if (ferror (fp)) {
			perror (argv[argn]);
			exit (2);
		}
		if (fp != stdin) {
			fclose (fp);
		}
	} else if (strcmp (arg, "-s") == 0) {
		otto_sched_t sched;
		if (argn+1 < argn) {
			usage ();
		}
		sched.length = 0;
		if (ioctl (fd, OTTO_GET_SCHED, (char *) &sched) == -1) {
			perror ("OTTO_GET_SCHED");
			exit (2);
		}
		if (argc - argn > 0) {
			setvc (fd, argc-argn, &argv[argn], 0, sched.length);
		}
	} else if (strcmp (arg, "-d") == 0) {
		unsigned def;
		if (ioctl (fd, OTTO_GET_DEFVC, (char *) &def) == -1) {
			perror ("OTTO_GET_DEFVC");
			exit (2);
		}
		while (argn != argc) {
			delete_mapping (fd, argv[argn++], def);
		}
	} else if (strcmp (arg, "-x") == 0) {
		otto_sched_t sched;
		unsigned char buf[1024];
		unsigned table[SCHEDSIZE];
		int i;

		sched.length = SCHEDSIZE;
		sched.table = table;

		if (argn == argc) {
			if (ioctl (fd, OTTO_GET_SCHED, (char *) &sched) == -1) {
				perror ("OTTO_GET_SCHED");
				exit (2);
			}
			for (i = 0; i != sched.length; i++) {
				printf ("%u\n", table[i]);
			}
		} else if (argn+1 == argc) {
			FILE *fp;
			if (strcmp (argv[argn], "-") == 0) {
				fp = stdin;
			} else {
				fp = fopen (argv[argn], "r");
			}
			if (fp == NULL) {
				perror (argv[argn]);
				exit (2);
			}
			sched.length = 0;
			while (sched.length != SCHEDSIZE &&
			       fgets (buf, sizeof (buf), fp) != NULL) {
				table[sched.length++] = 
					strtoul (buf, (char **) 0,0);
			}
			if (sched.length != 0) {
				if (ioctl (fd, OTTO_SET_SCHED,
						(char *)&sched)==-1) {
					perror("OTTO_SET_SCHED");
					exit (2);
				}
			}
			if (fp != stdin) {
				fclose (fp);
			}
		} else {
			usage ();
		}
	} else {
		usage ();
	}
	(void) close (fd);
	exit (0);
}
